#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <inttypes.h>

int64_t timespecDiff(struct timespec *timeA_p, struct timespec *timeB_p) {
  return ((timeA_p->tv_sec * 1000000) + (timeA_p->tv_nsec/1000)) -
           ((timeB_p->tv_sec * 1000000) + (timeB_p->tv_nsec/1000));
}

#ifndef DEFAULT_SIZE
#ifdef PERF
#define DEFAULT_SIZE 1024
#else
#define DEFAULT_SIZE 6
#endif
#endif

void mat_trans_opt(int *mat, int size) {
  //struct timespec start, end;

  register int s,x,y;
  register int *mat_src;
  int *mat_src_tmp;
  register int *mat_dst;
  int *mat_dst_tmp;

  //clock_gettime(CLOCK_MONOTONIC, &start);

  mat_src_tmp = mat + (size-1)*size;
  mat_dst_tmp = mat + (size-1);

  for ( x = size-1; mat_src_tmp > mat; mat_src_tmp -= size, mat_dst_tmp--, x-- ) {
    mat_dst = mat_dst_tmp;
    mat_src = mat_src_tmp;
    for ( y = 0; y < x; y++, mat_src++, mat_dst += size) {
      s = *mat_dst;
      *mat_dst = *mat_src;
      *mat_src = s;
    }
  }
  //clock_gettime(CLOCK_MONOTONIC, &end);
  //printf("Time trans: %" PRId64 "\n", (timespecDiff(&end, &start))/1000);

}

void mat_trans(int *mat, int size) {
  int s,x,y;

  for ( x = size-1; x > 0; x--) {
    for ( y = 0; y < x; y++) {
      s = mat[x*size+y];
      mat[x*size+y] = mat[y*size+x];
      mat[y*size+x] = s;
    }
  }
}

#ifdef PERF
#define BLOCKSIZE 8
#else
#define BLOCKSIZE 4
#endif

void mat_mul_opt_7(int *mat_1, int *mat_2, int *mat_res, int s) {
  mat_trans_opt(mat_2, s);


  int *mat_res_i = mat_res;        //y * size + x

  int *mat_1_s_b;
  int *mat_2_s_b;

  int x_b,y_b;

  int *mat_1_s = mat_1 + s - 1; //start
  int *mat_2_s;

  int x,y;

  register int *mat_1_p;
  register int *mat_2_p;

  register int res;
  register int size = s;

  for ( y = 0; y < size; y+=BLOCKSIZE, mat_1_s += BLOCKSIZE*size ) {

    mat_2_s = mat_2;

    for ( x = 0; x < size; x+=BLOCKSIZE, mat_2_s += BLOCKSIZE*size ) {

      mat_1_s_b = mat_1_s;

      for ( y_b = 0; y_b < BLOCKSIZE; y_b+=2, mat_1_s_b+=(size+size)) {

        mat_2_s_b = mat_2_s;

        for ( x_b = 0; x_b < BLOCKSIZE; x_b+=2, mat_2_s_b+=(size+size)) {

          mat_1_p = mat_1_s_b;
          mat_2_p = mat_2_s_b + size - 1;

          res = 0;

          for (; mat_2_p >= mat_2_s_b; mat_1_p--, mat_2_p--) {
             res += (*mat_1_p) * (*mat_2_p);
          }

          *(mat_res_i++) = res;

          res = 0;

          mat_1_p = mat_1_s_b;
          mat_2_p = mat_2_s_b + size + size - 1;

          for (; mat_2_p >= mat_2_s_b + size; mat_1_p--, mat_2_p--) {
             res += (*mat_1_p) * (*mat_2_p);
          }

          *(mat_res_i--) = res;

          mat_res_i+=size;

          mat_1_p = mat_1_s_b + size;
          mat_2_p = mat_2_s_b + size - 1;

          res = 0;

          for (; mat_2_p >= mat_2_s_b; mat_1_p--, mat_2_p--) {
             res += (*mat_1_p) * (*mat_2_p);
          }

          *(mat_res_i++) = res;

          mat_1_p = mat_1_s_b + size;
          mat_2_p = mat_2_s_b + size + size - 1;

          res = 0;

          for (; mat_2_p >= mat_2_s_b + size; mat_1_p--, mat_2_p--) {
             res += (*mat_1_p) * (*mat_2_p);
          }

          *(mat_res_i++) = res;

          mat_res_i-=size;

        }
        mat_res_i+=(size+size-BLOCKSIZE);
      }
      mat_res_i+=(BLOCKSIZE-(BLOCKSIZE*size));
    }
    mat_res_i+=size;
  }

  mat_trans_opt(mat_2, size);

}

#ifdef PERF
#define BLOCKSIZE 8
#else
#define BLOCKSIZE 2
#endif

void mat_mul_opt_6(int *mat_1, int *mat_2, int *mat_res, int size) {
  mat_trans_opt(mat_2, size);

  int *mat_1_p;
  int *mat_2_p;

  int *mat_res_i = mat_res;        //y * size + x

  int *mat_1_s_b;
  int *mat_2_s_b;

  int x_b,y_b;

  int *mat_1_s = mat_1 + size - 1; //start
  int *mat_2_s;

  int x,y;

  register int res;
  register int res_2;
  register int res_3;
  register int res_4;

  for ( y = 0; y < size; y+=BLOCKSIZE, mat_1_s += BLOCKSIZE*size ) {

    mat_2_s = mat_2;

    for ( x = 0; x < size; x+=BLOCKSIZE, mat_2_s += BLOCKSIZE*size ) {

      mat_1_s_b = mat_1_s;

      for ( y_b = 0; y_b < BLOCKSIZE; y_b+=2, mat_1_s_b+=(size+size)) {

        mat_2_s_b = mat_2_s;

        for ( x_b = 0; x_b < BLOCKSIZE; x_b+=2, mat_2_s_b+=(size+size)) {

          mat_1_p = mat_1_s_b;
          mat_2_p = mat_2_s_b + size - 1;

          res_4 = res_3 = res_2 = res = 0;

          //printf("Start: %d %d\n", (int)(mat_1_p-mat_1-size+1),(int)(mat_2_p-mat_2-size+1));

          for (; mat_2_p >= mat_2_s_b; mat_1_p--, mat_2_p--) {
             //printf("Mat_1: %d %d Mat_2: %d %d\n", (int)(mat_1_p - mat_1), (int)(mat_1_p - mat_1 + size), (int)(mat_2_p - mat_2), (int)(mat_2_p - mat_2 + size));
             res += (*mat_1_p) * (*mat_2_p);
             res_2 += (*mat_1_p) * (*(mat_2_p+size));
             res_4 += (*(mat_1_p+size)) * (*(mat_2_p+size));
             res_3 += (*(mat_1_p+size)) * (*mat_2_p);
          }

          //printf("P: %d %d\n  %d %d\n", (int)(mat_res_i - mat_res), (int)(mat_res_i - mat_res + 1), (int)(mat_res_i - mat_res + size), (int)(mat_res_i - mat_res + size + 1));
          *(mat_res_i+size) = res_3;
          *(mat_res_i++) = res;
          *(mat_res_i+size) = res_4;
          *(mat_res_i++) = res_2;
        }
        mat_res_i+=(size+size-BLOCKSIZE);
      }
      mat_res_i+=(BLOCKSIZE-(BLOCKSIZE*size));
    }
    mat_res_i+=size;
  }

  mat_trans_opt(mat_2, size);

}


#undef BLOCKSIZE
#ifdef PERF
#define BLOCKSIZE 8
#else
#define BLOCKSIZE 2
#endif

void mat_mul_opt_5(int *mat_1, int *mat_2, int *mat_res, int size) {
  mat_trans_opt(mat_2, size);

  int *mat_res_i = mat_res;        //y * size + x

  int *mat_1_s_b;
  int *mat_2_s_b;

  int x_b,y_b;

  int *mat_1_s = mat_1 + size - 1; //start
  int *mat_2_s;

  int x,y;

  register int res;
  register int res_2;

  register int *mat_1_p;
  register int *mat_2_p;

  for ( y = 0; y < size; y+=BLOCKSIZE, mat_1_s += BLOCKSIZE*size ) {

    mat_2_s = mat_2;

    for ( x = 0; x < size; x+=BLOCKSIZE, mat_2_s += BLOCKSIZE*size ) {

      mat_1_s_b = mat_1_s;

      for ( y_b = 0; y_b < BLOCKSIZE; y_b++, mat_1_s_b+=size) {

        mat_2_s_b = mat_2_s;

        for ( x_b = 0; x_b < BLOCKSIZE; x_b+=2, mat_2_s_b+=(size+size)) {

          mat_1_p = mat_1_s_b;
          mat_2_p = mat_2_s_b + size - 1;

          for ( res_2 = res = 0; mat_2_p >= mat_2_s_b; mat_1_p--, mat_2_p--) {
             res += (*mat_1_p) * (*mat_2_p);
             res_2 += (*mat_1_p) * (*(mat_2_p+size));
          }

          *(mat_res_i++) = res;
          *(mat_res_i++) = res_2;
        }
        mat_res_i+=(size-BLOCKSIZE);
      }
      mat_res_i+=(BLOCKSIZE-(BLOCKSIZE*size));
    }
    mat_res_i+=size;
  }

  mat_trans_opt(mat_2, size);

}

#undef BLOCKSIZE

#ifdef PERF
#define BLOCKSIZE 1024
#else
#define BLOCKSIZE 4
#endif


void mat_mul_opt_4(int *mat_1, int *mat_2, int *mat_res, int size) {
  mat_trans_opt(mat_2, size);

  int x,y,x_b,y_b;
  register int res;

  int *mat_res_i = mat_res; //y * size + x

  int *mat_1_s = mat_1 + size - 1; //start
  int *mat_2_s;

  int *mat_1_s_b;
  register int *mat_2_s_b;

  register int *mat_1_p;
  register int *mat_2_p;

  for ( y = 0; y < size; y+=BLOCKSIZE, mat_1_s += BLOCKSIZE*size ) {

    mat_2_s = mat_2;

    for ( x = 0; x < size; x+=BLOCKSIZE, mat_2_s += BLOCKSIZE*size ) {

      mat_1_s_b = mat_1_s;

      for ( y_b = 0; y_b < BLOCKSIZE; y_b++, mat_1_s_b+=size) {

        mat_2_s_b = mat_2_s;

        for ( x_b = 0; x_b < BLOCKSIZE; x_b++, mat_2_s_b+=size) {

          mat_1_p = mat_1_s_b;
          mat_2_p = mat_2_s_b + size - 1;

          for ( res = 0; mat_2_p >= mat_2_s_b; mat_1_p--, mat_2_p-- )
             res += (*mat_1_p) * (*mat_2_p);

          //printf("P: %d\n", (int)(mat_res_i - mat_res));
          *mat_res_i = res;
          mat_res_i++;
        }
        mat_res_i+=(size-BLOCKSIZE);
        //printf("Point end\n");
      }
      //Block end
      mat_res_i+=(BLOCKSIZE-(BLOCKSIZE*size));
      //printf("Block end\n");
    }
    //line end
    //printf("Blockline end: %d -> %d\n",(int)(mat_res_i - mat_res),(int)((mat_res_i+BLOCKSIZE-size) - mat_res));
    mat_res_i+=size;
  }

  mat_trans_opt(mat_2, size);

}

void mat_mul_opt_3(int *mat_1, int *mat_2, int *mat_res, int size) {
  mat_trans_opt(mat_2, size);

  int x,y;
  register int res;

  int res_i = 0; //y * size + x

  int *mat_1_s = mat_1 + size - 1; //start
  register int *mat_2_s;

  register int *mat_1_p;
  register int *mat_2_p;

  for ( y = 0; y < size*size; y+=size, mat_1_s += size ) {

    mat_2_s = mat_2;

    for ( x = 0; x < size; x++, res_i++, mat_2_s += size ) {

      mat_1_p = mat_1_s;
      mat_2_p = mat_2_s + size - 1;

      for ( res = 0; mat_2_p >= mat_2_s; mat_1_p--, mat_2_p-- )
         res += (*mat_1_p) * (*mat_2_p);

     mat_res[res_i] = res;

    }
  }

  mat_trans_opt(mat_2, size);

}

void mat_mul_opt_2(int *mat_1, int *mat_2, int *mat_res, int size) {
  int x, y, y_x, x2, x2_size_x;

  int res;
  int *mat_1_p;
  int *mat_2_p;

  for ( x = 0; x < size; x++ ) {

    mat_1_p = mat_1;

    for ( y = 0; y < size*size; y+=size ) {

      res = 0;

      mat_2_p = mat_2 + x;

      for ( x2 = 0; x2 < size; x2++, mat_1_p++, mat_2_p += size )
        res += (*mat_1_p) * (*mat_2_p);

      mat_res[y + x] = res;
    }
  }
}

void mat_mul_opt(int *mat_1, int *mat_2, int *mat_res, int size) {
  int x, y, y_x, x2, x2_size_x;

  for ( x = 0; x < size; x++ ) {
    for ( y = 0; y < size*size; y+=size ) {
      y_x = y + x;
      mat_res[y_x] = 0;
      for ( x2_size_x = x, x2 = 0; x2 < size; x2++, x2_size_x += size ) {
        mat_res[y_x] += mat_1[y + x2] * mat_2[x2_size_x];
      }
    }
  }
}

void mat_mul(int *mat_1, int *mat_2, int *mat_res, int size) {
  int x,y,x2;

  for ( x = 0; x < size; x++ ) {
    for ( y = 0; y < size; y++ ) {
      mat_res[y * size + x] = 0;
      for ( x2 = 0; x2 < size; x2++ ) {
        mat_res[y * size + x] += mat_1[y*size + x2] * mat_2[x2*size + x];
      }
    }
  }
}

int mat_test(int *mat_1, int *mat_2, int *mat_test, int size) {
  int *mat_res = (int*)malloc(size*size*sizeof(int));

  mat_mul(mat_1, mat_2, mat_res,size);

  int result = mat_cmp(mat_res, mat_test, size);

  free(mat_res);

  return result;
}

int mat_cmp(int *mat_1, int *mat_2, int size) {
  return memcmp(mat_1, mat_2, size*size);
}

void mat_print(int *mat, int size) {
  int x,y;
  for ( y = 0; y < size; y++ ) {
    for ( x = 0; x < size; x++ ) {
      printf("%d ",mat[y * size + x]);
    }
    printf("\n");
  }
  printf("\n");
}

int main(int argc, char **argv) {
  struct timespec start, end;
//#define TEST
  int r;
#ifndef TEST
  int size = DEFAULT_SIZE;
  int i;

  int *mat_1 = (int*)malloc(size*size*sizeof(int));
  int *mat_2 = (int*)malloc(size*size*sizeof(int));

  for ( i = 0; i < size*size; i++) {
    mat_1[i] = i%5;
    mat_2[i] = i%7;
  }

#else
  int size = 4;
  int mat_1[] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};
  int mat_2[] = { 4, 3, 2, 1, 8, 7, 6, 5, 12, 11, 10, 9, 16, 15, 14, 13};
#endif

  int *mat_res = (int*)malloc(size*size*sizeof(int));

  clock_gettime(CLOCK_MONOTONIC, &start);

  for ( r = 0; r < 1; r++) {
    if ( argc == 1)  mat_mul(mat_1, mat_2, mat_res, size);
    else {
      int v = atoi(argv[1]);
      switch (v) {
        case 1:
          mat_mul_opt(mat_1, mat_2, mat_res, size);
          break;
        case 2:
          mat_mul_opt_2(mat_1, mat_2, mat_res, size);
          break;
        case 3:
          mat_mul_opt_3(mat_1, mat_2, mat_res, size);
          break;
        case 4:
          mat_mul_opt_4(mat_1, mat_2, mat_res, size);
          break;
        case 5:
          mat_mul_opt_5(mat_1, mat_2, mat_res, size);
          break;
        case 6:
          mat_mul_opt_6(mat_1, mat_2, mat_res, size);
          break;
        case 7:
          mat_mul_opt_7(mat_1, mat_2, mat_res, size);
          break;
      }
    }
  }

  clock_gettime(CLOCK_MONOTONIC, &end);

#ifdef TEST
  mat_print(mat_1,size);
  mat_print(mat_2,size);
  mat_print(mat_res,size);
#endif

#ifndef PERF
  printf("Testresult: %d\n", mat_test(mat_1, mat_2, mat_res, size));
#endif

  printf("Time: %" PRId64 "\n", (timespecDiff(&end, &start))/1000);

#ifndef TEST
  free(mat_1);
  free(mat_2);
#endif
  free(mat_res);
}
